@@ -5,27 +5,35 @@ Helper classes for reading cached objects from Git's Object Database.
 from __future__ import annotations
 
 import hashlib
-import re
 import os
+import re
+import sys
+from collections import defaultdict
+from contextlib import AbstractContextManager
+from enum import Enum
+from pathlib import Path
+from subprocess import DEVNULL, PIPE, CalledProcessError, Popen, run
+from tempfile import NamedTemporaryFile, TemporaryDirectory
+from types import TracebackType
 from typing import (
-    TypeVar,
-    Type,
+    IO,
+    TYPE_CHECKING,
     Dict,
-    Union,
-    Sequence,
-    Optional,
-    Mapping,
     Generic,
+    Mapping,
+    Optional,
+    Sequence,
     Tuple,
+    Type,
+    TypeVar,
+    Union,
     cast,
 )
-import sys
-from types import TracebackType
-from pathlib import Path
-from enum import Enum
-from subprocess import Popen, run, PIPE, CalledProcessError
-from collections import defaultdict
-from tempfile import TemporaryDirectory
+
+if TYPE_CHECKING:
+    from subprocess import _FILE
+
+    from typing_extensions import Self
 
 
 class MissingObject(Exception):
@@ -53,7 +61,7 @@ class Oid(bytes):
     def __new__(cls, b: bytes) -> Oid:
         if len(b) != 20:
             raise ValueError("Expected 160-bit SHA1 hash")
-        return super().__new__(cls, b)  # type: ignore
+        return super().__new__(cls, b)
 
     @classmethod
     def fromhex(cls, instr: str) -> Oid:
@@ -161,8 +169,8 @@ class Repository:
     """path to GnuPG binary"""
 
     _objects: Dict[int, Dict[Oid, GitObj]]
-    _catfile: Popen
-    _tempdir: Optional[TemporaryDirectory]
+    _catfile: Popen[bytes]
+    _tempdir: Optional[TemporaryDirectory[str]]
 
     __slots__ = [
         "workdir",
@@ -217,10 +225,10 @@ class Repository:
         self,
         *cmd: str,
         cwd: Optional[Path] = None,
+        env: Optional[Dict[str, str]] = None,
         stdin: Optional[bytes] = None,
+        stdout: _FILE = PIPE,
         trim_newline: bool = True,
-        env: Dict[str, str] = None,
-        nocapture: bool = False,
     ) -> bytes:
         if cwd is None:
             cwd = getattr(self, "workdir", None)
@@ -228,17 +236,16 @@ class Repository:
         cmd = ("git",) + cmd
         prog = run(
             cmd,
-            stdout=None if nocapture else PIPE,
             cwd=cwd,
             env=env,
             input=stdin,
+            stdout=stdout,
             check=True,
         )
 
-        if nocapture:
-            return b""
-        if trim_newline and prog.stdout.endswith(b"\n"):
-            return prog.stdout[:-1]
+        if trim_newline and isinstance(prog.stdout, bytes):
+            if prog.stdout.endswith(b"\n"):
+                return prog.stdout[:-1]
         return prog.stdout
 
     def config(self, setting: str, default: T) -> Union[bytes, T]:
@@ -325,25 +332,66 @@ class Repository:
         key_id = self.config(
             "user.signingKey", default=self.default_committer.signing_key
         )
-        gpg = None
-        try:
-            gpg = sh_run(
-                (self.gpg, "--status-fd=2", "-bsau", key_id),
-                stdout=PIPE,
-                stderr=PIPE,
-                input=buffer,
-                check=True,
+        signer = None
+        if self.config("gpg.format", "gpg") == b"ssh":
+            program = self.config("gpg.ssh.program", b"ssh-keygen")
+            is_literal_ssh_key = key_id.startswith(b"ssh-") or key_id.startswith(
+                b"key::"
             )
-        except CalledProcessError as gpg:
-            print(gpg.stderr.decode(), file=sys.stderr, end="")
-            print("gpg failed to sign commit", file=sys.stderr)
-            raise
+            if is_literal_ssh_key and key_id.startswith(b"key::"):
+                key_id = key_id[5:]
+            if is_literal_ssh_key:
+                key_file_context_manager: AbstractContextManager[IO[bytes]] = (
+                    NamedTemporaryFile(  # pylint: disable=consider-using-with
+                        prefix=".git_signing_key_tmp"
+                    )
+                )
+            else:
+                key_file_context_manager = open(key_id, "rb")
+            with key_file_context_manager as key_file:
+                if is_literal_ssh_key:
+                    key_file.write(key_id)
+                    key_file.flush()
+                    key_id = key_file.name.encode("utf-8")
+                try:
+                    args = [program, "-Y", "sign", "-n", "git", "-f", key_id]
+                    if is_literal_ssh_key:
+                        args.append("-U")
+                    signer = sh_run(
+                        args, stdout=PIPE, stderr=PIPE, input=buffer, check=True
+                    )
+                except CalledProcessError as ssh:
+                    e = ssh.stderr.decode()
+                    print(e, file=sys.stderr, end="")
+                    print(f"{program.decode()} failed to sign commit", file=sys.stderr)
+                    if "usage:" in e:
+                        print(
+                            (
+                                "ssh-keygen -Y sign is needed for ssh signing "
+                                "(available in openssh version 8.2p1+)"
+                            ),
+                            file=sys.stderr,
+                        )
+                    raise
+        else:
+            try:
+                signer = sh_run(
+                    (self.gpg, "--status-fd=2", "-bsau", key_id),
+                    stdout=PIPE,
+                    stderr=PIPE,
+                    input=buffer,
+                    check=True,
+                )
+            except CalledProcessError as gpg:
+                print(gpg.stderr.decode(), file=sys.stderr, end="")
+                print("gpg failed to sign commit", file=sys.stderr)
+                raise
 
-        if b"\n[GNUPG:] SIG_CREATED " not in gpg.stderr:
-            raise GPGSignError(gpg.stderr.decode())
+            if b"\n[GNUPG:] SIG_CREATED " not in signer.stderr:
+                raise GPGSignError(signer.stderr.decode())
 
         signature = b"gpgsig"
-        for line in gpg.stdout.splitlines():
+        for line in signer.stdout.splitlines():
             signature += b" " + line + b"\n"
         return signature
 
@@ -477,7 +525,7 @@ class GitObj:
 
     __slots__ = ("repo", "body", "oid", "persisted")
 
-    def __new__(cls: Type[GitObjT], repo: Repository, body: bytes) -> GitObjT:
+    def __new__(cls, repo: Repository, body: bytes) -> "Self":
         oid = Oid.for_object(cls._git_type(), body)
         cache = repo._objects[oid[0]]  # pylint: disable=protected-access
         if oid in cache:
@@ -492,7 +540,7 @@ class GitObj:
         self.persisted = False
         cache[oid] = self
         self._parse_body()  # pylint: disable=protected-access
-        return cast(GitObjT, self)
+        return self
 
     @classmethod
     def _git_type(cls) -> str:
@@ -559,7 +607,7 @@ class Commit(GitObj):
 
         # Parse the header to populate header metadata fields.
         self.parent_oids = []
-        for hdr in re.split(br"\n(?! )", hdrs):
+        for hdr in re.split(rb"\n(?! )", hdrs):
             # Parse out the key-value pairs from the header, handling
             # continuation lines.
             key, value = hdr.split(maxsplit=1)
@@ -769,13 +817,24 @@ class Tree(GitObj):
         entry in the index will have its "Skip Workdir" bit set."""
 
         index = Index(self.repo, path)
-        self.repo.git("read-tree", "--index-output=" + str(path), self.persist().hex())
+        self.repo.git(
+            "read-tree",
+            "--index-output=" + str(path),
+            self.persist().hex(),
+            stdout=DEVNULL,
+        )
 
         # If skip_worktree is set, mark every file as --skip-worktree.
         if skip_worktree:
             # XXX(nika): Could be done with a pipe, which might improve perf.
             files = index.git("ls-files")
-            index.git("update-index", "--skip-worktree", "--stdin", stdin=files)
+            index.git(
+                "update-index",
+                "--skip-worktree",
+                "--stdin",
+                stdin=files,
+                stdout=DEVNULL,
+            )
 
         return index
 
@@ -814,21 +873,21 @@ class Index:
         self,
         *cmd: str,
         cwd: Optional[Path] = None,
-        stdin: Optional[bytes] = None,
-        trim_newline: bool = True,
         env: Optional[Mapping[str, str]] = None,
-        nocapture: bool = False,
+        stdin: Optional[bytes] = None,
+        stdout: _FILE = PIPE,
+        trim_newline: bool = True,
     ) -> bytes:
         """Invoke git with the given index as active"""
-        env = dict(**env) if env is not None else dict(**os.environ)
+        env = {**env} if env is not None else {**os.environ}
         env["GIT_INDEX_FILE"] = str(self.index_file)
         return self.repo.git(
             *cmd,
             cwd=cwd,
-            stdin=stdin,
-            trim_newline=trim_newline,
             env=env,
-            nocapture=nocapture,
+            stdin=stdin,
+            stdout=stdout,
+            trim_newline=trim_newline,
         )
 
     def tree(self) -> Tree:
@@ -870,7 +929,20 @@ class Reference(Generic[GitObjT]):  # pylint: disable=unsubscriptable-object
 
     def __init__(self, obj_type: Type[GitObjT], repo: Repository, name: str) -> None:
         self._type = obj_type
-        self.name = repo.git("rev-parse", "--symbolic-full-name", name).decode()
+
+        self.name = name
+        try:
+            # Silently verify that a ref with the name exists and recover if it
+            # doesn't.
+            repo.git("show-ref", "--quiet", "--verify", self.name)
+        except CalledProcessError:
+            # `name` could be a branch name which can be resolved to a ref. Try
+            # to do so with `rev-parse`, and verify that the new name exists.
+            self.name = repo.git(
+                "rev-parse", "--symbolic-full-name", self.name
+            ).decode()
+            repo.git("show-ref", "--verify", self.name)
+
         self.repo = repo
         self.refresh()
 
@@ -896,5 +968,5 @@ class Reference(Generic[GitObjT]):  # pylint: disable=unsubscriptable-object
         if self.target is not None:
             args.append(str(self.target.oid))
 
-        self.repo.git(*args)
+        self.repo.git(*args, stdout=DEVNULL)
         self.target = new
